----------------------------------------------------------------------------------
-- Engineer1: Programmcode Ralph Hnsel	Stand 02.10.2011 TFT_KOMBI fuer	AC1
-- Engineer2: BWS 64/80	Color Heiko	Poppe 29.09.2013
-- HW-Grundlage: NANOS Color-BWS
-- Module Name:	BWS_64_80_COLOR
-- Taktgenerator 25,175	Mhz	und	20,00 MHz
-- Zeichengenerator	6x8	fuer AC1-Mode und 8x8 (8x10) fuer 80x24-Mode
-- Ausgabe mit 800x600@60Hz	fuer 64x32 und 640x480@60Hz	fuer 80x24 Mode
-- Engineer3: Anpassungen fr den NKC Jens Mewes 18.10.2018
-- Farberweiterung entfernt, 1.Pixelreihe (64x80) gefixt, eine IO-Adresse (0x7h)
----------------------------------------------------------------------------------
library	IEEE;
use	IEEE.STD_LOGIC_1164.ALL;
use	IEEE.NUMERIC_STD.ALL ; --notwendig fr die Zaehler !

entity BWS_64_80_COLOR is
   Port	( 
			DB			: inout	STD_LOGIC_VECTOR (3 downto 0);		--	4	-- Datenbus Eingang/Ausgang
			ADR7		: in	STD_LOGIC_VECTOR (7 downto 0);		--	8	-- Adresse	A7..A0
			ADR			: in	STD_LOGIC_VECTOR (15 downto 11);	--	5	-- Adresse A15..A11
			ADRC		: in	STD_LOGIC_VECTOR (15 downto 11);	--	5	-- A15..A11 Vorgabe fuer	BWS-Adressbereich
			ADRC7		: in	STD_LOGIC_VECTOR (7 downto 4);		--	4	-- A7..A4 Vorgabe fuer IO-Adressbereich
			IORQ		: in	STD_LOGIC;							--	1	--
			MREQ		: in	STD_LOGIC;							--	1	--
			RD			: in	STD_LOGIC;							--	1	--
			WR			: in	STD_LOGIC;							--	1	--
			Reset		: in	STD_LOGIC;							--	1	--
			MEMDI_IN	: in	STD_LOGIC;							--	1	-- L deaktiviert den BWS
			MEMDI_OUT	: out	STD_LOGIC;							--	1	-- L wenn	der	BWS	aktiv ist
			BWS_Mode	: inout	STD_LOGIC;							--	1	-- L = AC1-Mode (64x32), H = 80x24, an A12 ZG
			MOD64_80	: in	STD_LOGIC;							--	1	-- Start mit 64 Zeichen (L) oder 80 Zeichen (H)
			AC1_ADR		: in	STD_LOGIC;							--	1	-- L = Adressen im 64x32-Mode	invertiert (wie	AC1)
			RAMADR		: out	STD_LOGIC_VECTOR (10 downto 0);		--	11	-- A10..A0 BWS AusleseRAM	(links)
		 	ZGADR		: out	STD_LOGIC_VECTOR (2 downto 0);		--	3	-- A2..A0 ZG (Zeilenadresse	Zeichen) 
			ZGADR11		: out	STD_LOGIC;							--	1	-- A11 Zeichengenerator
			CSRAM		: out	STD_LOGIC;							--	1	-- CS	Signal fr RAM (rechts)	CPU-Seite
--			CSFARBRAM	: out	STD_LOGIC;							--	1	-- CS	Signal fr FarbRAM (rechts)	CPU-Seite
--			CSFARBLATCH	: out	STD_LOGIC;							--	1	-- Datenuebernahme	Latch aus FarbRAM
			ZG_Out		: in	STD_LOGIC_VECTOR (7 downto 0);		--	8	-- Ausgaenge ZG (intern an Schieberegister)
			DTVDU		: inout	STD_LOGIC;							--	1	-- Dunkeltastung Bildraender extern
			Video_out	: out	STD_LOGIC;							--	1	-- VideoSignal
			HK_Sync		: out	STD_LOGIC;							--	1	-- HK SynchronImpuls  negiert (wegen Treiber)
			VK_Sync		: out	STD_LOGIC;							--	1	-- VK SynchronImpuls  negiert (wegen Treiber)
			Takt_in		: in	STD_LOGIC;							--	1	-- Eingangstakt	25,175 MHz
			Takt2in		: in	STD_LOGIC							--	1	-- Eingangstakt 2 20 /	40 MHz
			);														---------------------------------------------------
																	-- 68 Pins

end	BWS_64_80_COLOR;

architecture VERHALTEN of BWS_64_80_COLOR is
	signal	VDCLK		: STD_LOGIC	:= '0';										-- VDCLK = Bildpunkttakt
	signal	QD19		: UNSIGNED (2 downto 0) := (others	=> '0');			-- Ausgang D19 8:1 Taktteiler
	signal	QD41_42		: UNSIGNED (7 downto 0) := (others	=> '0');			-- Ausgang D41 und 42 Zeichenzaehler
	signal	QTFT		: UNSIGNED (0 downto 0) := (others	=> '0');			-- Ausgang QTFT	2:1	Teiler fr TFT Modus
	signal	QD43_44_20	: UNSIGNED (8 downto 0) := (others	=> '0');			-- Ausgang D43,D44,D20 Zeilenzaehler
	signal	QADDR		: UNSIGNED (10 downto 0) := (others	=> '0');			-- Zaehler fuer	BWS-RAM-Adresse
	signal	QZEILE		: UNSIGNED (3 downto 0) := (others	=> '0');			-- Zaehler fuer	eine Zeichenzeile
	signal	ADRLATCH1	: UNSIGNED (10 downto 0) := (others	=> '0');			-- Zwischenspeicher	Zeilenwiederholung
	signal	BWS_Zugriff	: STD_LOGIC	:= '0';										-- Signal 1	fr	Zugriff	der	CPU	auf	BWS
	signal	SR_ZG_Out	: STD_LOGIC_VECTOR (7 downto 0) := (others	=> '0');	-- Ausgaenge Schieberegister ZG
	signal	DG13		: STD_LOGIC	:= '0';										-- Dunkeltastung = 1 bei Seitenrand	L+R
	signal	DG14		: STD_LOGIC	:= '0';										-- Dunkeltastung = 1 bei Bildrand O+U
	signal	DG15		: STD_LOGIC	:= '0';										-- Dunkeltastung zwischen den Zeilen =1	(NANOS-Mode)
	signal	StartZeile	: STD_LOGIC	:= '0';										-- Signal 1	Start neue Zeile
	signal	IO12_Latch	: STD_LOGIC_VECTOR (3 downto 0)	:= (others => '0');		-- I/O - Register mit 4	Bit	Breite
-- **************************************************************************************************************

begin

-- hier	das	zentrale Taktsignal	fr	alle Prozesse definieren

	VDCLK	<= Takt2in	
			when BWS_Mode = '0'		-- Eingangstakt	2 (20,00 MHz fuer SVGA-Mode)
			else Takt_In ;			-- Eingangstakt	1 (25,175 MHz fuer VGA-Mode)


-- IO-Register mit 4 Bit Breite	(DB3..0)		Bit0 - BWS Speicher: L=Eingeblendet, H=Ausgeblendet
--												Bit1 - Umschaltung 64/80 Zeichen: L=64,	H=80
--												Bit2 - L = Dunkeltastung des Zeilenzwischenraums bei 80x24
--												BIT3 - L = Zeichensatz 1, H	= Zeichensatz 2	fuer Mode 64x32	
process(Takt2in)
begin
	if rising_edge(Takt2in)	then
		if (ADR7(7 downto 4) = ADRC7(7 downto 4) and IORQ =	'0'	and	WR = '0') then
			if	ADR7(3 downto 0) = x"7"	then IO12_Latch(3 downto 0)	 <=	DB(3 downto	0) ;	-- IO Adresse 0X1h Steuerung
			--elsif ADR7(3 downto	0) = x"2" then IO12_Latch(0)  <= DB(0) ;					-- IO Adresse 0X2h Color-Umschaltung
			end	if ;
		elsif (Reset = '0' and MOD64_80	= '0') then	IO12_Latch	<= "0000" ;					-- Reset, Start	im 64Z Modus
		elsif (Reset = '0' and MOD64_80	= '1') then	IO12_Latch	<= "0010" ;					-- Reset, Start	im 80Z Modus
		else IO12_Latch	 <=	IO12_Latch ;													-- Zustand behalten			
		end	if;		
	end	if;	
end	Process;

-- Signal BWS_Mode aus Bit1	des	IO12_LATCH,	-> geht	an A12 des ZG
	BWS_Mode <=	IO12_Latch(1);

--	IO12H Register auslesen, Daten auf DB	
	DB <= IO12_LATCH	when (ADR7(7 downto 4) = ADRC7(7 downto 4) 
							and	(ADR7(3	downto 0)) = x"7"					-- Adresse 0X7h	zum	Auslesen
							and	IORQ = '0'									-- IORQ	aktiv
							and	RD = '0')									-- IO-Register ruecklesen
						else "ZZZZ";										-- DB hochohmig
		
--	Adressdekodierung &	Zugriff	selektieren	
	BWS_Zugriff <= '1'	when ADR(15 downto 11) = ADRC(15 downto 11)			-- Adressen	vergleichen
							and	MREQ = '0'									-- Speicherzugriff
							--and IO12_Latch(0) = '0'							-- wenn	BWS	aktiv sein darf
							and ((IO12_Latch(0) = '0') or (MEMDI_IN = '0'))	-- wenn	BWS	aktiv sein darf
						else '0';


	
-- Memdi-Signal	fuer externen RAM (alternativ zu MEMDI_IN)
	MEMDI_OUT <= not BWS_Zugriff;

-- CS fr BWS RAM erzeugen	
	CSRAM <= '0'	when BWS_Zugriff = '1'				-- normaler	BWS-Zugriff
						--and	IO12_Latch(0) =	'0'			-- FarbRAM inaktiv
					else '1';							-- RAM BWS inaktiv

-- CS fr Farb RAM erzeugen		
--	CSFARBRAM <= '0'	when BWS_Zugriff = '1'			-- normaler	BWS-Zugriff
--							and	IO12_Latch(0) =	'1'		-- FarbRAM aktiv
--						else '1';						-- RAM Farbe inaktiv

--	Daten vom Farb-RAM ins Farb_Latch einlesen
--process (VDCLK)
--begin	
--	if rising_edge(VDCLK) then								-- hier	RISING-EDGE	sonst flimmern		
--		if QD19	= 0												-- Timing Zeichen /	Hintergrund	auf	0			
--			then	CSFARBLATCH	<=	'1'	;			
--			else  CSFARBLATCH <= '0' ;					-- Taktsignal fr Latch	74LS374	L/H	speichert		
--		end	if;		
--	end	if;	
--end	process;

-- ZG Daten	ins	Schieberegister	SR_ZG_Out holen	und	herausschieben,	
ZG_Out_ausschieben:		
process	(VDCLK)						
	begin	
		if falling_edge(VDCLK) then		  
			if QD19 =	0 then	SR_ZG_Out(7	downto 0) <= ZG_Out(7 downto 0)	;	-- parallel	laden	 Breite	8 Pixel
			else SR_ZG_Out <= SR_ZG_Out(6 downto 0)	& '0' ;							-- Schieberichtung fuer	NANOS-ZG
--			else SR_ZG_Out <= '0' &	SR_ZG_Out(7	downto 1) ;							-- Schieberichtung fuer	AC1-ZG
		end	if;
	end if;
end process ZG_Out_ausschieben;

-- Video-Ausgang	-- Ausgang Schieberegister an Video_Out
--	Video_Out <= SR_ZG_out(0) and not DTVDU	;										-- Schieberichtung fuer	AC1-ZG
	Video_Out <= not (SR_ZG_out(7) and not DTVDU and not DG15) ;			-- Schieberichtung fuer	NANOS-ZG	

-- Signal DG13 - Seitenraender links + rechts
process	(VDCLK)
begin
	if	falling_edge(VDCLK)	then
		if		(BWS_MODE = '0' and QD41_42 = 65	and  QD19 = 0) then DG13 <= '1' ;			-- Dunkeltastung bei 64x32 Zeichen-Modus
		elsif	(BWS_MODE = '1' and QD41_42 = 80)	then DG13 <= '1' ;			-- Dunkeltastung bei 80x24 Zeichen-Modus
		elsif	(BWS_Mode =	'0'	and	QD41_42	= 1)	then DG13 <= '0' ;			-- Anzeige aktiv 64x32
		elsif	(BWS_Mode = '1' and QD41_42 = 0)	then DG13 <= '0' ;			-- Anzeige aktiv 80x24
		end	if;
	end	if;
end	process;

-- Signal DG14 - Bildraender oben und unten
process	(VDCLK)
begin
	if	falling_edge(VDCLK)	then
		if	   (BWS_Mode = '0' and QD43_44_20 =	256) then DG14 <= '1' ;	-- Dunkeltastung bei 64x32 Zeichen-Modus
		elsif  (BWS_Mode = '1' and QD43_44_20 =	240) then DG14 <= '1' ;	-- Dunkeltastung bei 80x24 Zeichen-Modus
		elsif  QD43_44_20 =	0  then	DG14 <=	'0'	;								-- Anzeige aktiv
		end	if;
	end	if;
end	process;

-- Signal DG15 - Dunkeltastung zwischen	den	Zeilen bei NANOS-Mode, nur wenn	IO12_Latch Bit2	= L
Process	(VDCLK)
begin
	if	falling_edge(VDCLK)	then
		if	   (IO12_Latch(2) =	'0'	and	QZEILE = 8 and BWS_Mode	= '1')
			then DG15 <= '1' ;															-- Dunkeltastung bei NANOS-Mode
		elsif  QZEILE =	0 then DG15	<= '0' ;										-- Anzeige aktiv
		end	if;
	end	if;
end	process;

-- Signal DTVDU	bilden.. (Dunkeltastsignal fr die Raender & Austastluecken)
	DTVDU	<= '1' when	 DG14 =	'1'	or DG13	= '1'									-- Ausgang zur Dunkeltastung
						else '0' ;			

-- ********	Zaehlerkette fuer BWS-Adressen ***********************************************

-- QZEILE -	Zaehler	fuer Zeichengenerator Bit 0	- 3	im 80Z Mode	
process	(VDCLK)		--	
begin	
	if rising_edge(VDCLK) then 
		if StartZeile =	'1'	then								-- Zeilenende (bei TFT-Mode	Ende 2.	Durchlauf)
			if (QZEILE < 9 and QD43_44_20 <	240)			-- nur im sichtbaren Zeilenbereich
				then QZEILE	   <= QZEILE + 1 ;				-- zaehlt 0..9,	8 Pixelzeilen (0..7) je	Zeichenzeile + 2 Leerzeilen					
				else QZEILE	   <= "0000";						-- sonst ruecksetzen
			end	if;			
		end	if;	
	end	if;
end	process;

-- ********	QADDR -	Adresszaehler fuer BWS-RAM im 80Z Mode *****************************************
process	(VDCLK)		--	
begin
	if rising_edge(VDCLK) then
--
-- 1. am Bildende Adresszaehler	ruecksetzen
		if (QD43_44_20 = 250)														-- letzte Zeile	-- 
			then QADDR <= "00000000000";											-- dann	ruecksetzen
		end	if;	

-- 2. am Bildende Adresslatch ruecksetzen (muss	mit	Adresszaehler korrespondieren)
		if QD43_44_20 =	250															-- letzte Zeile				
			then ADRLATCH1 <= "00000000000";										-- dann	ruecksetzen			
		end	if;	

-- 3. Am Ende einer	neuen Zeichenzeile die neue	BWS-Adresse	setzen
		if (QD19 = 4 and QZEILE	= 9	and	QD41_42	= 98 and QTFT =	1)		-- Zeilen- und Zeichenende
			then ADRLATCH1 <= ADRLATCH1	+ 80;									-- dann	neue Adresse setzen
		end	if;	

-- 4. am Zeilenende	innerhalb eines	Zeichens BS-Adresse	in Zaehler zurueckladen
		if (QD19 = 2 and QD41_42 = 96)											-- 
			then QADDR(10 downto 0)	<= UNSIGNED(ADRLATCH1(10 downto	0));
		end	if;

-- 5. Hochzaehlen innerhalb	der	Zeile, nur im sichtbaren Bildbereich
		if (QD19 = 3 and QD41_42 < 80 and QD43_44_20 < 240)				-- Takt	+ Freigabe 
			then QADDR <= QADDR	+ 1;												-- zaehlen
		end	if ;
--		
	end	if ;
end process;


-- RAM Adressen	bilden ------------------------------------------------
process(VDCLK)
begin
	if rising_edge(VDCLK) then
		if	  BWS_Mode = '1' then
				RAMADR(10 downto 0)	<= STD_LOGIC_VECTOR(QADDR(10 downto	0))	;				-- Adressen	fuer BWS-RAM 80	x 24 Mode
				ZGADR(2	downto 0)	<= STD_LOGIC_VECTOR(QZEILE(2 downto	0))		;			-- Bits	0..2 Zeilenzaehler
				ZGADR11				<= QZEILE(3)	;
		elsif	(BWS_Mode =	'0'	and	AC1_ADR	= '0') then												-- Mode	AC1	64 x 32					
				RAMADR(5 downto	0)	<= not STD_LOGIC_VECTOR(QD41_42(5 downto 0)) ;		-- mit negiertem Adressbus				
				RAMADR(10 downto 6)	<= not STD_LOGIC_VECTOR(QD43_44_20(7 downto	3))	;				
				ZGADR(2	downto 0)	<= STD_LOGIC_VECTOR(QD43_44_20(2 downto	0))	;				
				ZGADR11				<= IO12_Latch(3) ;												-- Umschaltung Zeichensatz		
		elsif	(BWS_Mode =	'0'	and	AC1_ADR	= '1') then	
				RAMADR(5 downto	0)	<= STD_LOGIC_VECTOR(QD41_42(5 downto 0)) ;
				RAMADR(10 downto 6)	<= STD_LOGIC_VECTOR(QD43_44_20(7 downto	3))	;
				ZGADR(2	downto 0)	<= STD_LOGIC_VECTOR(QD43_44_20(2 downto	0))	;				
				ZGADR11				<= IO12_Latch(3) ;		
		end	if ;
	end	if ;
end	process;

-- *********** Zaehlerkette	fuer Ablenkfrequenzen *******************************************
-- Die Zaehlerkette	erzeugt	
-- im 64x32	Modus ein SVGA-Signal 800x600@60Hz und
-- im 80x24	Modus ein  VGA-Signal 640x480@60Hz

-- Breite 64 Zeichen mit je	6 Pixel	bzw. 80	Zeichen	mit	je 8 Pixel,	Rest ist linker	und	rechter	Rand

-- D19 - Basis ist VDCLK mit 20,00 MHz (64Z) bzw. 25,175 MHz (80Z)
D19:
process (VDCLK)	 --	Takt wird durch	6 bzw. 8 geteilt = 6 oder 8	Spalten	je Zeichen	
begin	
	if rising_edge(VDCLK) then	
		if	  BWS_Mode = '1' then QD19 <= QD19 + 1 ;
		elsif (BWS_Mode	= '0' and QD19 < 5)	
			then QD19 <= QD19 +	1 ;
			else QD19 <= "000" ;
		end	if ;	
	end	if;	
end	process	D19;

-- Spaltenzhler D41_D42 zaehlt	Zeichen	0..100 (80Z) bzw. 0..87	(64Z), incl. Leerraeume	& HK-Synchronimpuls
D41_42:	
process	(VDCLK)
begin
	if rising_edge(VDCLK) then
		if (BWS_Mode = '0' and QD19	= 5)							-- fuer	Timing Hintergrund
		or (BWS_Mode = '1' and QD19	= 7) then
			if (BWS_Mode = '0' and QD41_42 < 87)					-- Mode	64x32
			or (BWS_Mode = '1' and QD41_42 < 99)					-- Mode	80x24
				then		QD41_42	  <= QD41_42 + 1 ;				-- zaehlen
				else		QD41_42	  <= "00000000"	 ; 
			end	if;	
		end	if;
	end	if ;
end process D41_42;				

TFT_Zeile2:		-- 2:1 Teiler sorgt	dafr, dass	jede Zeile 2x geschrieben wird
process	(VDCLK)
begin
	if rising_edge(VDCLK) then
		if QD19	= 2	then
			if	  (BWS_Mode	= '0' and QD41_42 =	86)				-- Mode	64x32
			or	  (BWS_Mode	= '1' and QD41_42 =	98)				-- Mode	80x24
			then	 QTFT <= QTFT +	1 ;
			end	if ;		
		end	if ;
	end	if ;
end process TFT_Zeile2;				

-- Hilfssignal Startzeile, bestimmt	wann Zeilenzaehler weiterschaltet	
StartZeile <= '1' when							
					(BWS_Mode =	'0'	and	QD19 = 3 and QD41_42 = 83 and QTFT = 1)		-- Mode	64x32
				or	(BWS_Mode =	'1'	and	QD19 = 3 and QD41_42 = 97 and QTFT = 1)		-- 2. Durchlauf							
					else '0' ;	

-- Zeilenzaehler D43 & D44 & D20 zaehlt	Zeilen 0 bis 313
process	(VDCLK)
begin	
	if rising_edge(VDCLK)  then	
		if StartZeile =	'1'	then
			if	  (BWS_Mode	= '0' and QD43_44_20 < 313)			-- Mode	64x32
			or	  (BWS_Mode	= '1' and QD43_44_20 < 261)		-- Mode	80x24
				then		QD43_44_20 <= QD43_44_20 + 1 ;			-- zaehlen..
				else		QD43_44_20 <= "000000000"	  ;			-- ruecksetzen
			end	if;
		end	if ;
	end	if ;	
end	process;

-- *** BWS Zaehlerkette	--Ende----------------------------------------------

-- ********	Erzeugung HK Synchronimpuls	aus	Zhlerkette	*******************
process	(VDCLK)
begin
	if rising_edge(VDCLK) then
		if	BWS_Mode = '0' then										-- Mode	64x32			
			if	   QD41_42 = 69	then HK_sync <=	'1';				-- 69 Start	HK Impuls			
			elsif  QD41_42 = 80	then HK_sync <=	'0';				-- 80 Ende HK Impuls		
			end	if;			
		elsif BWS_Mode = '1' then										-- Mode	80x24
			if	   QD41_42 = 80	then HK_sync <=	'1';				-- 80 Start	HK Impuls			
			elsif  QD41_42 = 92	then HK_sync <=	'0';				-- 92 Ende HK Impuls			
			end	if;			
		end	if;
	end	if;
end	process;

-- ***Ende HK_Sync Impuls ****************************************************

-- ********	Erzeugung VK Synchronimpuls	aus	Zhlerkette	**********************
process	(VDCLK)
begin
	if rising_edge(VDCLK) then
		if	  BWS_Mode = '0' then										-- Mode	64x32
			if	QD43_44_20	= 278 then VK_sync <= '1';			-- 278 Start VK	Impuls			
			elsif QD43_44_20	= 280 then VK_sync <= '0';		-- 280 Ende	VK Imuls			
			end	if;
		elsif BWS_Mode = '1' then										-- Mode	80x24
			if	QD43_44_20	= 240 then VK_sync <= '1';			-- 240 Start VK	Impuls			
			elsif QD43_44_20	= 241 then VK_sync <= '0';		-- 241 Ende	VK Imuls			
			end	if;
		end	if;
	end	if;	
end	process;

-- ***Ende VK_Sync Impuls ****************************************************

end	VERHALTEN;